#include <stdio.h>
#include <assert.h>
#include "PtrMacro.h"
#include "CacheBlock.h"
#include "MIPS.h"

#define GROWSIZE	(0x40000)

CCacheBlock::TARGET_CPU		CCacheBlock::m_nTargetCPU = TARGET_CPU_UNINIT;

CCacheBlock::CCacheBlock(uint32 nStart, uint32 nEnd)
{
	m_nStart	= nStart;
	m_nEnd		= nEnd;

	m_pData		= (uint8*)malloc(0);
	m_nSize		= 0;
	m_nPtr		= 0;

	m_pReloc	= (uint32*)malloc((nEnd - nStart));

	m_nValid			= false;
	m_nCheckDelayedJump	= false;

	GetTargetProcessor();
}

CCacheBlock::~CCacheBlock()
{
	DELETEPTR(m_pData);
	DELETEPTR(m_pReloc);
}

void CCacheBlock::GetTargetProcessor()
{
	uint32 nFeatures;

	if(m_nTargetCPU != TARGET_CPU_UNINIT) return;

	__asm
	{
		mov eax, 0x00000001;
		cpuid;
		mov dword ptr[nFeatures], edx
	}

	//Check SSE2
	if(nFeatures & 0x04000000)
	{
		m_nTargetCPU = TARGET_CPU_SSE2;
		return;
	}

	//Check SSE
	if(nFeatures & 0x02000000)
	{
		m_nTargetCPU = TARGET_CPU_SSE;
		return;
	}
}

RET_CODE CCacheBlock::Execute(CMIPS* pCtx)
{
	void* pEntry;
	RET_CODE nRet;
	uint32 nAddress;

	if(!m_nValid)
	{
		Compile(pCtx);
	}

	nAddress = pCtx->m_pAddrTranslator(pCtx, 0x00000000, pCtx->m_State.nPC);

	pEntry = (uint8*)m_pData + m_pReloc[(nAddress - m_nStart) / 4];

	__asm
	{
		push	ebx
		push	esi
		push	edi

		mov		eax, [pEntry];
		call	eax
		mov		[nRet], eax;

		pop		edi
		pop		esi
		pop		ebx
	}

	return nRet;
}

void CCacheBlock::Compile(CMIPS* pCtx)
{
	uint32 i;
	bool nJump;

	m_nCheckDelayedJump = pCtx->IsBranch(m_nStart - 4);

	//Check if there's a jump instruction before this one
	for(i = m_nStart; i < m_nEnd; i += 4)
	{
		m_pReloc[(i - m_nStart) / 4] = m_nPtr;

		m_nProgramCounterChanged = false;
		nJump = m_nCheckDelayedJump;

		InsertProlog(pCtx);
		pCtx->m_pArch->CompileInstruction(i, this, pCtx, true);
		InsertEpilog(pCtx, nJump);

	}
	//Check jump out of the block
	Ret(RET_CODE_INTERBLOCKJUMP);
	m_nValid = true;
	printf("CacheBlock: Compiled block [0x%0.8X, 0x%0.8X]. %i byte(s) allocated.\r\n", m_nStart, m_nEnd, m_nSize);
}

void CCacheBlock::InsertProlog(CMIPS* pCtx)
{
/*
	//Increment PC
	PushAddr(&pCtx->m_State.nPC);
	AddImm(4);
	PullAddr(&pCtx->m_State.nPC);
*/

	//mov eax, $pCtx->m_State.nPC
	StreamWriteByte(0xB8);
	StreamWriteWord((uint32)((uint8*)(&pCtx->m_State.nPC) - (uint8*)0));

	//add [eax], 4
	StreamWriteByte(0x83);
	StreamWriteByte(0x00);
	StreamWriteByte(0x04);
}

void CCacheBlock::InsertEpilog(CMIPS* pCtx, bool nDelayJump)
{
	//////////////////////////////
	//Decrement quota

/*
	PushAddr(&pCtx->m_nQuota);
	SubImm(1);
	PullAddr(&pCtx->m_nQuota);
*/

	//mov eax, $pCtx->nQuota
	StreamWriteByte(0xB8);
	StreamWriteWord((uint32)((uint8*)(&pCtx->m_nQuota) - (uint8*)0));

	//dec [eax]
	StreamWriteByte(0xFF);
	StreamWriteByte(0x08);

	//////////////////////////////

	//Check delay slot (we truly need to do something better about this...)
	if(nDelayJump)
	{
		m_nCheckDelayedJump = false;

		PushAddr(&pCtx->m_State.nDelayedJumpAddr);
		PushImm(MIPS_INVALID_PC);
		Cmp(JCC_CONDITION_EQ);
		BeginJcc(false);
		{
			PushAddr(&pCtx->m_State.nDelayedJumpAddr);
			PullAddr(&pCtx->m_State.nPC);
		}
		EndJcc();

		//Reset the delay slot address
		PushImm(MIPS_INVALID_PC);
		PullAddr(&pCtx->m_State.nDelayedJumpAddr);

		//PC Changed
		SetProgramCounterChanged();
	}

	//Call the tick function (push a different value than 1 if there's a dead loop in there)
	PushImm(1);
	Call(pCtx->m_pTickFunction, 1, true);
	BeginJcc(true);
	{
		Ret(RET_CODE_BREAKPOINT);
	}
	EndJcc();

	//Check if it's 0
	PushAddr(&pCtx->m_nQuota);
	PushImm(0);
	Cmp(JCC_CONDITION_EQ);
	BeginJcc(true);
	{
		Ret(RET_CODE_QUOTADONE);
	}
	EndJcc();

	//Take the jump
	if(m_nProgramCounterChanged)
	{
		SynchornizePC(pCtx);
	}

}

void CCacheBlock::SynchornizePC(CMIPS* pCtx)
{
	//Check if we can take the execution back inside our block
	PushAddr(&pCtx->m_State.nPC);
	PushImm(m_nStart);
	Cmp(JCC_CONDITION_BL);

	BeginJcc(true);
	{
		Ret(RET_CODE_INTERBLOCKJUMP);
	}
	EndJcc();

	PushAddr(&pCtx->m_State.nPC);
	PushImm(m_nEnd);
	Cmp(JCC_CONDITION_BL);

	BeginJcc(false);
	{
		Ret(RET_CODE_INTERBLOCKJUMP);
	}
	EndJcc();

	//This will go to the PC
	PushAddr(&m_pData);
	PushAddr(&m_pReloc);
	
	PushAddr(&pCtx->m_State.nPC);
	SubImm(m_nStart);
	
	Add();

	//pop eax
	StreamWriteByte(0x58);

	//pop ebx
	StreamWriteByte(0x5B);

	//mov eax, [eax]
	StreamWriteByte(0x8B);
	StreamWriteByte(0x00);

	//add eax, ebx
	StreamWriteByte(0x03);
	StreamWriteByte(0xC3);

	//jmp eax
	StreamWriteByte(0xFF);
	StreamWriteByte(0xE0);
}

///////////////////////////////////////////////////////////
//Stream Fonction Implementation
///////////////////////////////////////////////////////////

void CCacheBlock::StreamWriteByte(uint8 nByte)
{
	if(m_nPtr == m_nSize)
	{
		m_nSize += GROWSIZE;
		m_pData = (uint8*)realloc(m_pData, m_nSize);
	}
	m_pData[m_nPtr] = nByte;
	m_nPtr++;
}

void CCacheBlock::StreamWriteWord(uint32 nWord)
{
	unsigned int i;
	for(i = 0; i < 4; i++)
	{
		StreamWriteByte(((uint8*)&nWord)[i]);
	}
}

///////////////////////////////////////////////////////////
//Intermediate Language Implementation
///////////////////////////////////////////////////////////

void CCacheBlock::PushAddr(void* pAddr)
{
	//push [pAddr]
	StreamWriteByte(0xFF);
	StreamWriteByte(0x35);
	StreamWriteWord((uint32)((uint8*)pAddr - (uint8*)0));
}

void CCacheBlock::PullAddr(void* pAddr)
{
	//pop [pAddr]
	StreamWriteByte(0x8F);
	StreamWriteByte(0x05);
	StreamWriteWord((uint32)((uint8*)pAddr - (uint8*)0));
}

void CCacheBlock::PushImm(uint32 nValue)
{
	//push nValue
	StreamWriteByte(0x68);
	StreamWriteWord(nValue);
}

void CCacheBlock::PushRef(void* pRef)
{
	PushImm((uint32)((uint8*)pRef - (uint8*)0));
}

void CCacheBlock::SeX8()
{
	//pop eax
	StreamWriteByte(0x58);

	//movsx eax, al
	StreamWriteByte(0x0F);
	StreamWriteByte(0xBE);
	StreamWriteByte(0xC0);

	//push eax
	StreamWriteByte(0x50);
}

void CCacheBlock::SeX16()
{
	//pop eax
	StreamWriteByte(0x58);

	//cwde
	StreamWriteByte(0x98);

	//push eax
	StreamWriteByte(0x50);
}

void CCacheBlock::SeX32()
{
	//pop eax
	StreamWriteByte(0x58);

	//cdq
	StreamWriteByte(0x99);

	//push eax
	StreamWriteByte(0x50);

	//push edx
	StreamWriteByte(0x52);
}

void CCacheBlock::PullTop()
{
	Free(1);
}

void CCacheBlock::Free(unsigned int nAmount)
{
	nAmount *= 4;

	assert(nAmount < 128);

	//add esp, 4
	StreamWriteByte(0x83);
	StreamWriteByte(0xC4);
	StreamWriteByte((uint8)nAmount);
}

void CCacheBlock::PushTop()
{
	//push [esp]
	StreamWriteByte(0xFF);
	StreamWriteByte(0x34);
	StreamWriteByte(0x24);
}

void CCacheBlock::PushValueAt(unsigned int nAddress)
{
	nAddress *= 4;
	
	assert(nAddress < 128);

	//push [esp + nAddress];
	StreamWriteByte(0xFF);
	StreamWriteByte(0x74);
	StreamWriteByte(0x24);
	StreamWriteByte((uint8)nAddress);
}

void CCacheBlock::Lookup(uint32* pTable)
{
	//pop eax
	StreamWriteByte(0x58);

	//mov ebx, $pTable
	StreamWriteByte(0xBB);
	StreamWriteWord((uint32)((uint8*)pTable - (uint8*)0));

	//mov eax, [eax * 4 + ebx]
	StreamWriteByte(0x8B);
	StreamWriteByte(0x04);
	StreamWriteByte(0x83);

	//push eax
	StreamWriteByte(0x50);
}

void CCacheBlock::Exchange(unsigned int nA1, unsigned int nA2)
{
	nA1 *= 4;
	nA2 *= 4;

	assert(nA1 < 128);
	assert(nA2 < 128);

	//mov eax, [esp + nA1]
	StreamWriteByte(0x8B);
	StreamWriteByte(0x44);
	StreamWriteByte(0x24);
	StreamWriteByte((uint8)nA1);

	//xchg eax, [esp + nA2]
	StreamWriteByte(0x87);
	StreamWriteByte(0x44);
	StreamWriteByte(0x24);
	StreamWriteByte((uint8)nA2);

	//mov [esp + nA1], eax
	StreamWriteByte(0x89);
	StreamWriteByte(0x44);
	StreamWriteByte(0x24);
	StreamWriteByte((uint8)nA1);
}

void CCacheBlock::Swap()
{
	Exchange(0, 1);
}

void CCacheBlock::And()
{
	//pop eax
	StreamWriteByte(0x58);

	//and [esp], eax
	StreamWriteByte(0x21);
	StreamWriteByte(0x04);
	StreamWriteByte(0x24);
}

void CCacheBlock::AndImm(uint32 nValue)
{
	//and [esp], nValue
	StreamWriteByte(0x81);
	StreamWriteByte(0x24);
	StreamWriteByte(0x24);
	StreamWriteWord(nValue);
}

void CCacheBlock::Add()
{
	//pop eax
	StreamWriteByte(0x58);

	//add [esp], eax
	StreamWriteByte(0x01);
	StreamWriteByte(0x04);
	StreamWriteByte(0x24);
}

void CCacheBlock::AddC()
{
	//pop eax
	StreamWriteByte(0x58);

	//pop ebx
	StreamWriteByte(0x5B);

	//push 0
	StreamWriteByte(0x6A);
	StreamWriteByte(0x00);

	//add ebx, eax
	StreamWriteByte(0x03);
	StreamWriteByte(0xD8);

	//adc [esp], 0
	StreamWriteByte(0x80);
	StreamWriteByte(0x14);
	StreamWriteByte(0x24);
	StreamWriteByte(0x00);

	//push ebx
	StreamWriteByte(0x53);
}

void CCacheBlock::AddImm(uint32 nValue)
{
	//add [esp], nValue
	StreamWriteByte(0x81);
	StreamWriteByte(0x04);
	StreamWriteByte(0x24);
	StreamWriteWord(nValue);
}

void CCacheBlock::Sub()
{
	//pop eax
	StreamWriteByte(0x58);

	//sub [esp], eax
	StreamWriteByte(0x29);
	StreamWriteByte(0x04);
	StreamWriteByte(0x24);
}

void CCacheBlock::SubImm(uint32 nValue)
{
	//sub [esp], nValue
	StreamWriteByte(0x81);
	StreamWriteByte(0x2C);
	StreamWriteByte(0x24);
	StreamWriteWord(nValue);
}

void CCacheBlock::SubC()
{
	//pop eax
	StreamWriteByte(0x58);

	//pop ebx
	StreamWriteByte(0x5B);

	//push 0
	StreamWriteByte(0x6A);
	StreamWriteByte(0x00);

	//sub ebx, eax
	StreamWriteByte(0x2B);
	StreamWriteByte(0xD8);

	//adc [esp], 0
	StreamWriteByte(0x80);
	StreamWriteByte(0x14);
	StreamWriteByte(0x24);
	StreamWriteByte(0x00);

	//push ebx
	StreamWriteByte(0x53);
}

void CCacheBlock::PushConditionBit(JCC_CONDITION nCond)
{
	switch(nCond)
	{
	case JCC_CONDITION_EQ:
		//je
		StreamWriteByte(0x74);
		break;
	case JCC_CONDITION_NE:
		//jne
		StreamWriteByte(0x75);
		break;
	case JCC_CONDITION_BL:
		//jb
		StreamWriteByte(0x72);
		break;
	case JCC_CONDITION_BE:
		//jbe
		StreamWriteByte(0x76);
		break;
	case JCC_CONDITION_LT:
		//jl
		StreamWriteByte(0x7C);
		break;
	case JCC_CONDITION_LE:
		//jle
		StreamWriteByte(0x7E);
		break;
	default:
		assert(0);
		break;
	}

	//$04 (label for previous jump)
	StreamWriteByte(0x04);

	//push 0
	StreamWriteByte(0x6A);
	StreamWriteByte(0x00);

	//jmp $02
	StreamWriteByte(0xEB);
	StreamWriteByte(0x02);

	//push 1
	StreamWriteByte(0x6A);
	StreamWriteByte(0x01);
}

void CCacheBlock::Cmp(JCC_CONDITION nCond)
{
	//pop eax
	StreamWriteByte(0x58);

	//cmp [esp], eax
	StreamWriteByte(0x39);
	StreamWriteByte(0x04);
	StreamWriteByte(0x24);

	//pop eax
	StreamWriteByte(0x58);

	PushConditionBit(nCond);
}

void CCacheBlock::Call(void* pFunc, unsigned int nParams, bool nSaveRes)
{
	//mov eax, pFunc
	StreamWriteByte(0xB8);
	StreamWriteWord((uint32)((uint8*)pFunc - (uint8*)0));

	//call eax
	StreamWriteByte(0xFF);
	StreamWriteByte(0xD0);

	if(nParams != 0)
	{
		//add esp, [nParams * 4];
		StreamWriteByte(0x83);
		StreamWriteByte(0xC4);
		StreamWriteByte(nParams * 4);
	}

	if(nSaveRes)
	{
		//push eax
		StreamWriteByte(0x50);
	}
}

void CCacheBlock::BeginJcc(bool nCond)
{
	//pop eax
	StreamWriteByte(0x58);

	//test eax, eax
	StreamWriteByte(0x85);
	StreamWriteByte(0xC0);

	if(!nCond)
	{
		//jne $l
		StreamWriteByte(0x75);
	}
	else
	{
		//je $l
		StreamWriteByte(0x74);
	}

	m_nJccPos = m_nPtr;
	StreamWriteByte(0x00);
}

void CCacheBlock::EndJcc()
{
	uint8 nDist;
	nDist = (uint8)(m_nPtr - (m_nJccPos + 1));
	m_pData[m_nJccPos] = nDist;
}

void CCacheBlock::Ret(RET_CODE nCode)
{
	//mov eax, nCode
	StreamWriteByte(0xB8);
	StreamWriteWord(nCode);

	//ret (stack unwinding?)
	StreamWriteByte(0xC3);
}

void CCacheBlock::Mult()
{
	//pop eax
	StreamWriteByte(0x58);

	//pop ebx
	StreamWriteByte(0x5B);

	//mul ebx
	StreamWriteByte(0xF7);
	StreamWriteByte(0xE3);

	//push edx
	StreamWriteByte(0x52);

	//push eax
	StreamWriteByte(0x50);
}

void CCacheBlock::MultS()
{
	//pop eax
	StreamWriteByte(0x58);

	//pop ebx
	StreamWriteByte(0x5B);

	//imul ebx
	StreamWriteByte(0xF7);
	StreamWriteByte(0xEB);

	//push edx
	StreamWriteByte(0x52);

	//push eax
	StreamWriteByte(0x50);
}

void CCacheBlock::Div()
{
	//pop ebx
	StreamWriteByte(0x5B);

	//pop eax
	StreamWriteByte(0x58);

	//xor edx, edx
	StreamWriteByte(0x33);
	StreamWriteByte(0xD2);

	//div ebx
	StreamWriteByte(0xF7);
	StreamWriteByte(0xF3);

	//push edx
	StreamWriteByte(0x52);

	//push eax
	StreamWriteByte(0x50);
}

void CCacheBlock::DivS()
{
	//pop ebx
	StreamWriteByte(0x5B);

	//pop eax
	StreamWriteByte(0x58);

	//cdq
	StreamWriteByte(0x99);

	//idiv ebx
	StreamWriteByte(0xF7);
	StreamWriteByte(0xFB);

	//push edx
	StreamWriteByte(0x52);

	//push eax
	StreamWriteByte(0x50);
}

void CCacheBlock::Or()
{
	//pop eax
	StreamWriteByte(0x58);

	//or [esp], eax
	StreamWriteByte(0x09);
	StreamWriteByte(0x04);
	StreamWriteByte(0x24);
}

void CCacheBlock::OrImm(uint32 nValue)
{
	//or [esp], nValue
	StreamWriteByte(0x81);
	StreamWriteByte(0x0C);
	StreamWriteByte(0x24);
	StreamWriteWord(nValue);
}

void CCacheBlock::Xor()
{
	//pop eax
	StreamWriteByte(0x58);

	//xor [esp], eax
	StreamWriteByte(0x31);
	StreamWriteByte(0x04);
	StreamWriteByte(0x24);
}

void CCacheBlock::Not()
{
	//not [esp]
	StreamWriteByte(0xF7);
	StreamWriteByte(0x14);
	StreamWriteByte(0x24);
}

void CCacheBlock::Sll()
{
	//pop ecx
	StreamWriteByte(0x59);

	//shl [esp], cl
	StreamWriteByte(0xD3);
	StreamWriteByte(0x24);
	StreamWriteByte(0x24);
}

void CCacheBlock::SllImm(uint8 nAmount)
{
	//shl [esp], nAmount
	StreamWriteByte(0xC1);
	StreamWriteByte(0x24);
	StreamWriteByte(0x24);
	StreamWriteByte(nAmount);
}

void CCacheBlock::Sll64()
{
	//pop ecx
	StreamWriteByte(0x59);

	//pop edx
	StreamWriteByte(0x5A);

	//pop eax
	StreamWriteByte(0x58);

	//test cl, cl
	StreamWriteByte(0x84);
	StreamWriteByte(0xC9);

	//je $done
	StreamWriteByte(0x74);
	StreamWriteByte(0x20);

	//cmp cl, 0x40
	StreamWriteByte(0x80);
	StreamWriteByte(0xF9);
	StreamWriteByte(0x40);

	//jae $zero
	StreamWriteByte(0x73);
	StreamWriteByte(0x17);

	//cmp cl, 0x20
	StreamWriteByte(0x80);
	StreamWriteByte(0xF9);
	StreamWriteByte(0x20);

	//jae $more32
	StreamWriteByte(0x73);
	StreamWriteByte(0x07);

	//shld edx, eax, cl
	StreamWriteByte(0x0F);
	StreamWriteByte(0xA5);
	StreamWriteByte(0xC2);

	//shl eax, cl
	StreamWriteByte(0xD3);
	StreamWriteByte(0xE0);

	//jmp $done
	StreamWriteByte(0xEB);
	StreamWriteByte(0x0F);

	//$more32

	//mov edx, eax
	StreamWriteByte(0x8B);
	StreamWriteByte(0xD0);
	
	//xor eax, eax
	StreamWriteByte(0x33);
	StreamWriteByte(0xC0);

	//and cl, 0x1F
	StreamWriteByte(0x80);
	StreamWriteByte(0xE1);
	StreamWriteByte(0x1F);

	//shl edx, cl
	StreamWriteByte(0xD3);
	StreamWriteByte(0xE2);

	//jmp $done
	StreamWriteByte(0xEB);
	StreamWriteByte(0x04);

	//$zero

	//xor eax, eax
	StreamWriteByte(0x33);
	StreamWriteByte(0xC0);

	//xor edx, edx
	StreamWriteByte(0x33);
	StreamWriteByte(0xD2);

	//$done

	//push eax
	StreamWriteByte(0x50);

	//push edx
	StreamWriteByte(0x52);
}

void CCacheBlock::Srl()
{
	//pop ecx
	StreamWriteByte(0x59);

	//shr [esp], cl
	StreamWriteByte(0xD3);
	StreamWriteByte(0x2C);
	StreamWriteByte(0x24);
}

void CCacheBlock::SrlImm(uint8 nAmount)
{
	//shr [esp], nAmount
	StreamWriteByte(0xC1);
	StreamWriteByte(0x2C);
	StreamWriteByte(0x24);
	StreamWriteByte(nAmount);
}

void CCacheBlock::Srl64()
{
	//pop ecx
	StreamWriteByte(0x59);

	//pop edx
	StreamWriteByte(0x5A);

	//pop eax
	StreamWriteByte(0x58);

	//test cl, cl
	StreamWriteByte(0x84);
	StreamWriteByte(0xC9);

	//je $done
	StreamWriteByte(0x74);
	StreamWriteByte(0x20);

	//cmp cl, 0x40
	StreamWriteByte(0x80);
	StreamWriteByte(0xF9);
	StreamWriteByte(0x40);

	//jae $zero
	StreamWriteByte(0x73);
	StreamWriteByte(0x17);

	//cmp cl, 0x20
	StreamWriteByte(0x80);
	StreamWriteByte(0xF9);
	StreamWriteByte(0x20);

	//jae $more32
	StreamWriteByte(0x73);
	StreamWriteByte(0x07);

	//shrd eax, edx, cl
	StreamWriteByte(0x0F);
	StreamWriteByte(0xAD);
	StreamWriteByte(0xD0);

	//shr edx, cl
	StreamWriteByte(0xD3);
	StreamWriteByte(0xEA);

	//jmp $done
	StreamWriteByte(0xEB);
	StreamWriteByte(0x0F);

	//$more32

	//mov eax, edx
	StreamWriteByte(0x8B);
	StreamWriteByte(0xC2);
	
	//xor edx, edx
	StreamWriteByte(0x33);
	StreamWriteByte(0xD2);

	//and cl, 0x1F
	StreamWriteByte(0x80);
	StreamWriteByte(0xE1);
	StreamWriteByte(0x1F);

	//shr eax, cl
	StreamWriteByte(0xD3);
	StreamWriteByte(0xE8);

	//jmp $done
	StreamWriteByte(0xEB);
	StreamWriteByte(0x04);

	//$zero

	//xor eax, eax
	StreamWriteByte(0x33);
	StreamWriteByte(0xC0);

	//xor edx, edx
	StreamWriteByte(0x33);
	StreamWriteByte(0xD2);

	//$done

	//push eax
	StreamWriteByte(0x50);

	//push edx
	StreamWriteByte(0x52);

}

void CCacheBlock::Sra()
{
	//pop ecx
	StreamWriteByte(0x59);

	//sar [esp], cl
	StreamWriteByte(0xD3);
	StreamWriteByte(0x3C);
	StreamWriteByte(0x24);
}

void CCacheBlock::SraImm(uint8 nAmount)
{
	//sar [esp], nAmount
	StreamWriteByte(0xC1);
	StreamWriteByte(0x3C);
	StreamWriteByte(0x24);
	StreamWriteByte(nAmount);
}

void CCacheBlock::SetDelayedJumpCheck()
{
	m_nCheckDelayedJump = true;
}

void CCacheBlock::SetProgramCounterChanged()
{
	m_nProgramCounterChanged = true;
}

///////////////////////////////////////////////////////////
//FPU Intermediate Language Implementation
///////////////////////////////////////////////////////////

void CCacheBlock::FPU_PushAddrSingle(void* pAddr)
{
	//mov eax, $pAddr
	StreamWriteByte(0xB8);
	StreamWriteWord((uint32)((uint8*)pAddr - (uint8*)0));

	//fld dword [eax]
	StreamWriteByte(0xD9);
	StreamWriteByte(0x00);
}

void CCacheBlock::FPU_PushAddrWord(void* pAddr)
{
	//mov eax, $pAddr
	StreamWriteByte(0xB8);
	StreamWriteWord((uint32)((uint8*)pAddr - (uint8*)0));

	//fild dword [eax]
	StreamWriteByte(0xDB);
	StreamWriteByte(0x00);
}

void CCacheBlock::FPU_PullAddrSingle(void* pAddr)
{
	//mov eax, $pAddr
	StreamWriteByte(0xB8);
	StreamWriteWord((uint32)((uint8*)pAddr - (uint8*)0));

	//fstp dword [eax]
	StreamWriteByte(0xD9);
	StreamWriteByte(0x18);
}

void CCacheBlock::FPU_PullAddrWord(void* pAddr)
{
	//mov eax, $pAddr
	StreamWriteByte(0xB8);
	StreamWriteWord((uint32)((uint8*)pAddr - (uint8*)0));

	//fistp dword [eax]
	StreamWriteByte(0xDB);
	StreamWriteByte(0x18);
}

void CCacheBlock::FPU_Add()
{
	//faddp
	StreamWriteByte(0xDE);
	StreamWriteByte(0xC1);
}

void CCacheBlock::FPU_Sub()
{
	//fsubp
	StreamWriteByte(0xDE);
	StreamWriteByte(0xE9);
}

void CCacheBlock::FPU_Mul()
{
	//fmulp
	StreamWriteByte(0xDE);
	StreamWriteByte(0xC9);
}

void CCacheBlock::FPU_Div()
{
	//fdivp
	StreamWriteByte(0xDE);
	StreamWriteByte(0xF9);
}

void CCacheBlock::FPU_Neg()
{
	//fchs
	StreamWriteByte(0xD9);
	StreamWriteByte(0xE0);
}

void CCacheBlock::FPU_Cmp(JCC_CONDITION nCond)
{
	//fcomp
	StreamWriteByte(0xD8);
	StreamWriteByte(0xD9);

	//fstp st
	StreamWriteByte(0xDD);
	StreamWriteByte(0xD8);

	//wait
	StreamWriteByte(0x9B);

	//fnstsw ax
	StreamWriteByte(0xDF);
	StreamWriteByte(0xE0);

	//sahf
	StreamWriteByte(0x9E);

	PushConditionBit(nCond);
}

///////////////////////////////////////////////////////////
//VU Intermediate Language Implementation
///////////////////////////////////////////////////////////

void CCacheBlock::VU_SubB128()
{
	if(m_nTargetCPU == TARGET_CPU_SSE2)
	{
		//movdqu xmm0, [esp + 0x10]
		StreamWriteWord(0x446F0FF3);
		StreamWriteByte(0x24);
		StreamWriteByte(0x10);

		//movdqu xmm1, [esp + 0x00]
		StreamWriteWord(0x0C6F0FF3);
		StreamWriteByte(0x24);

		//add esp, 0x10
		StreamWriteByte(0x83);
		StreamWriteByte(0xC4);
		StreamWriteByte(0x10);

		//psubb xmm0, xmm1
		StreamWriteWord(0xC1F80F66);

		//movdqu [esp], xmm0
		StreamWriteWord(0x047F0FF3);
		StreamWriteByte(0x24);
	}
	else if(m_nTargetCPU == TARGET_CPU_SSE)
	{
		//movq mm0, [esp + 0x10]
		StreamWriteWord(0x24446F0F);
		StreamWriteByte(0x10);

		//movq mm1, [esp + 0x18]
		StreamWriteWord(0x244C6F0F);
		StreamWriteByte(0x18);

		//movq mm2, [esp + 0x00]
		StreamWriteWord(0x24146F0F);

		//movq mm3, [esp + 0x08]
		StreamWriteWord(0x245C6F0F);
		StreamWriteByte(0x08);

		//add esp, 0x10
		StreamWriteByte(0x83);
		StreamWriteByte(0xC4);
		StreamWriteByte(0x10);

		//psubb mm0, mm2
		StreamWriteByte(0x0F);
		StreamWriteByte(0xF8);
		StreamWriteByte(0xC2);

		//psubb mm1, mm3
		StreamWriteByte(0x0F);
		StreamWriteByte(0xF8);
		StreamWriteByte(0xCB);

		//movq [esp + 0x00], mm0
		StreamWriteWord(0x24047F0F);

		//movq [esp + 0x08], mm1
		StreamWriteWord(0x244C7F0F);
		StreamWriteByte(0x08);

		//emms
		StreamWriteByte(0x0F);
		StreamWriteByte(0x77);
	}
	else
	{
		assert(0);
	}
}

void CCacheBlock::VU_SubW128()
{
	if(m_nTargetCPU == TARGET_CPU_SSE2)
	{
		//movdqu xmm0, [esp + 0x10]
		StreamWriteWord(0x446F0FF3);
		StreamWriteByte(0x24);
		StreamWriteByte(0x10);

		//movdqu xmm1, [esp + 0x00]
		StreamWriteWord(0x0C6F0FF3);
		StreamWriteByte(0x24);

		//add esp, 0x10
		StreamWriteByte(0x83);
		StreamWriteByte(0xC4);
		StreamWriteByte(0x10);

		//psubd xmm0, xmm1
		StreamWriteWord(0xC1FA0F66);

		//movdqu [esp], xmm0
		StreamWriteWord(0x047F0FF3);
		StreamWriteByte(0x24);
	}
	else if(m_nTargetCPU == TARGET_CPU_SSE)
	{
		//movq mm0, [esp + 0x10]
		StreamWriteWord(0x24446F0F);
		StreamWriteByte(0x10);

		//movq mm1, [esp + 0x18]
		StreamWriteWord(0x244C6F0F);
		StreamWriteByte(0x18);

		//movq mm2, [esp + 0x00]
		StreamWriteWord(0x24146F0F);

		//movq mm3, [esp + 0x08]
		StreamWriteWord(0x245C6F0F);
		StreamWriteByte(0x08);

		//add esp, 0x10
		StreamWriteByte(0x83);
		StreamWriteByte(0xC4);
		StreamWriteByte(0x10);

		//psubd mm0, mm2
		StreamWriteByte(0x0F);
		StreamWriteByte(0xFA);
		StreamWriteByte(0xC2);

		//psubd mm1, mm3
		StreamWriteByte(0x0F);
		StreamWriteByte(0xFA);
		StreamWriteByte(0xCB);

		//movq [esp + 0x00], mm0
		StreamWriteWord(0x24047F0F);

		//movq [esp + 0x08], mm1
		StreamWriteWord(0x244C7F0F);
		StreamWriteByte(0x08);

		//emms
		StreamWriteByte(0x0F);
		StreamWriteByte(0x77);
	}
	else
	{
		assert(0);
	}
}

///////////////////////////////////////////////////////////
//Proxies Implementations
///////////////////////////////////////////////////////////

uint32 CCacheBlock::GetByteProxy(CMIPS* pCtx, uint32 nAddress)
{
	return (uint32)pCtx->m_pMemoryMap->GetByte(nAddress);
}

uint32 CCacheBlock::GetHalfProxy(CMIPS* pCtx, uint32 nAddress)
{
	return (uint32)pCtx->m_pMemoryMap->GetHalf(nAddress);
}

uint32 CCacheBlock::GetWordProxy(CMIPS* pCtx, uint32 nAddress)
{
	return pCtx->m_pMemoryMap->GetWord(nAddress);
}

void CCacheBlock::SetByteProxy(CMIPS* pCtx, uint32 nValue, uint32 nAddress)
{
	pCtx->m_pMemoryMap->SetByte(nAddress, (uint8)(nValue & 0xFF));
}

void CCacheBlock::SetHalfProxy(CMIPS* pCtx, uint32 nValue, uint32 nAddress)
{
	pCtx->m_pMemoryMap->SetHalf(nAddress, (uint16)(nValue & 0xFFFF));
}

void CCacheBlock::SetWordProxy(CMIPS* pCtx, uint32 nValue, uint32 nAddress)
{
	pCtx->m_pMemoryMap->SetWord(nAddress, nValue);
}
