#include <assert.h>
#include <string.h>
#include "IOP_FileIO.h"
#include "PS2VM.h"

using namespace IOP;
using namespace Framework;

CFileIO::CFileIO()
{
	m_Device.Insert(new CDirectoryDevice("mc0",		"./vfs/mc0"));
	m_Device.Insert(new CDirectoryDevice("host",	"./vfs/host"));
	//Insert standard files.
	m_File.Insert(new CStdioFile(fopen("ps2_stdout.txt", "ab")), 1);
	m_File.Insert(new CStdioFile(fopen("ps2_stderr.txt", "ab")), 2);
	m_nFileID = 3;
}

CFileIO::~CFileIO()
{
	while(m_File.Count() != 0)
	{
		delete m_File.Pull();
	}
	while(m_Device.Count() != 0)
	{
		delete m_Device.Pull();
	}
}

void CFileIO::Invoke(uint32 nMethod, void* pArgs, uint32 nArgsSize, void* pRet, uint32 nRetSize)
{
	switch(nMethod)
	{
	case 0x00000000:
		assert(nRetSize == 4);
		*(uint32*)pRet = Open(*(uint32*)pArgs, ((char*)pArgs) + 4);
		break;
	case 0x00000001:
		assert(nRetSize == 4);
		*(uint32*)pRet = Close(*(uint32*)pArgs);
		break;
	case 0x00000002:
		assert(nRetSize == 4);
		*(uint32*)pRet = Read(((uint32*)pArgs)[0], ((uint32*)pArgs)[2], (void*)(CPS2VM::m_pRAM + ((uint32*)pArgs)[1]));
		break;
	case 0x00000003:
		assert(nRetSize == 4);
		*(uint32*)pRet = Write(((uint32*)pArgs)[0], ((uint32*)pArgs)[2], (void*)(CPS2VM::m_pRAM + ((uint32*)pArgs)[1]));
		break;
	case 0x00000004:
		assert(nRetSize == 4);
		*(uint32*)pRet = Seek(((uint32*)pArgs)[0], ((uint32*)pArgs)[1], ((uint32*)pArgs)[2]);
		break;
	default:
		assert(0);
		break;
	}
}

bool CFileIO::SplitPath(const char* sPath, char* sDevice, char* sDevicePath)
{
	const char* sSplit;
	unsigned int nLenght;
	unsigned int nDevLenght;
	unsigned int nDevPathLenght;

	nLenght = (unsigned int)strlen(sPath);
	sSplit = strchr(sPath, ':');
	if(sSplit == NULL)
	{
		//Invalid path
		return false;
	}

	nDevLenght		= (unsigned int)(sSplit - sPath);
	nDevPathLenght	= (unsigned int)(nLenght - (sSplit - sPath) - 1);

	strncpy(sDevice, sPath, nDevLenght);
	strncpy(sDevicePath, sSplit + 1, nDevPathLenght);

	sDevice[nDevLenght] = '\0';
	sDevicePath[nDevPathLenght] = '\0';

	return true;
}

CFileIO::CDevice* CFileIO::FindDevice(const char* sDeviceName)
{
	CList<CDevice>::ITERATOR itDevice;
	CDevice* pDevice;

	for(itDevice = m_Device.Begin(); !itDevice; itDevice++)
	{
		pDevice = (*itDevice);
		if(!strcmp(pDevice->GetName(), sDeviceName))
		{
			return pDevice;
		}
	}

	return NULL;
}

uint32 CFileIO::RegisterFile(CFile* pFile)
{
	uint32 nFD;
	nFD = m_nFileID;
	m_File.Insert(pFile, nFD);
	m_nFileID++;
	return nFD;
}

uint32 CFileIO::Open(uint32 nMode, const char* sPath)
{
	char sDevice[256];
	char sDevicePath[256];
	CDevice* pDevice;
	CFile* pFile;

	printf("IOP_FileIO: Attempting to open file '%s'.\r\n", sPath);

	if(!SplitPath(sPath, sDevice, sDevicePath))
	{
		return 0xFFFFFFFF;
	}

	pDevice = FindDevice(sDevice);
	if(pDevice == NULL)
	{
		return 0xFFFFFFFF;
	}

	pFile = pDevice->OpenFile(nMode, sDevicePath);
	if(pFile == NULL)
	{
		return 0xFFFFFFFF;
	}

	return RegisterFile(pFile);
}

uint32 CFileIO::Close(uint32 nFile)
{
	CFile* pFile;

	pFile = m_File.Remove(nFile);
	if(pFile == NULL)
	{
		return 0xFFFFFFFF;
	}

	delete pFile;

	return 0;
}

uint32 CFileIO::Read(uint32 nFile, uint32 nSize, void* pBuffer)
{
	CFile* pFile;
	
	pFile = m_File.Find(nFile);
	if(pFile == NULL)
	{
		return 0xFFFFFFFF;
	}

	return pFile->Read(nSize, pBuffer);
}

uint32 CFileIO::Write(uint32 nFile, uint32 nSize, void* pBuffer)
{
	CFile* pFile;

	pFile = m_File.Find(nFile);
	if(pFile == NULL)
	{
		printf("IOP_FileIO: Writing to an unopened file (fd = %i).\r\n", nFile);
		return 0xFFFFFFFF;
	}

	return pFile->Write(nSize, pBuffer);
}

uint32 CFileIO::Seek(uint32 nFile, uint32 nOffset, uint32 nWhence)
{
	CFile* pFile;

	pFile = m_File.Find(nFile);
	if(pFile == NULL)
	{
		printf("IOP_FileIO: Seeking in an unopened file (fd = %i).\r\n", nFile);
		return 0xFFFFFFFF;
	}

	return pFile->Seek(nOffset, nWhence);
}

/////////////////////////////////////////////
//Devices Implementation
/////////////////////////////////////////////

CFileIO::CDevice::CDevice(const char* sName)
{
	m_sName = sName;
}

const char* CFileIO::CDevice::GetName()
{
	return m_sName;
}

CFileIO::CDirectoryDevice::CDirectoryDevice(const char* sName, const char* sBasePath) :
CDevice(sName)
{
	m_sBasePath = sBasePath;
}

CFileIO::CFile* CFileIO::CDirectoryDevice::OpenFile(uint32 nMode, const char* sDevicePath)
{
	FILE* pStream;
	char sPath[256];
	const char* sMode;

	strcpy(sPath, m_sBasePath);
	if(sDevicePath[0] != '/')
	{
		strcat(sPath, "/");
	}
	strcat(sPath, sDevicePath);

	switch(nMode)
	{
	case 0:
	case O_RDONLY:
		sMode = "rb";
		break;
	case (O_RDWR | O_CREAT):
		sMode = "w+";
		break;
	default:
		assert(0);
		break;
	}


	pStream = fopen(sPath, sMode);
	if(pStream == NULL) return NULL;

	return new CStdioFile(pStream);
}

/////////////////////////////////////////////
//Files Implementation
/////////////////////////////////////////////

CFileIO::CStdioFile::CStdioFile(FILE* pStream)
{
	m_pStream = pStream;
}

CFileIO::CStdioFile::~CStdioFile()
{
	if(m_pStream != NULL)
	{
		fclose(m_pStream);
	}
}

uint32 CFileIO::CStdioFile::Read(uint32 nLenght, void* pBuffer)
{
	return (uint32)fread(pBuffer, 1, nLenght, m_pStream);
}

uint32 CFileIO::CStdioFile::Write(uint32 nLenght, void* pBuffer)
{
	uint32 nCount;
	nCount = (uint32)fwrite(pBuffer, 1, nLenght, m_pStream);
	fflush(m_pStream);
	return nCount;
}

uint32 CFileIO::CStdioFile::Seek(uint32 nOffset, uint32 nWhence)
{
	int nRet;
	switch(nWhence)
	{
	case 0:
		nWhence = SEEK_SET;
		break;
	case 1:
		nWhence = SEEK_CUR;
		break;
	case 2:
		nWhence = SEEK_END;
		break;
	}
	nRet = fseek(m_pStream, nOffset, nWhence);
	if(nRet == -1) return nRet;
	return ftell(m_pStream);
}
